# If you have come till here, and you want to work on more things like this, reach out:
# research@nimblebox.ai

import os
import threading
from time import sleep
from json import dumps

import nbox.utils as U
from nbox.utils import logger, SimplerTimes
from nbox.init import nbox_grpc_stub
from nbox.auth import secret, AuthConfig, auth_info_pb
from nbox.hyperloop.jobs.job_pb2 import Job as JobProto
from nbox.hyperloop.jobs.dag_pb2 import Node
from nbox.hyperloop.jobs.nbox_ws_pb2 import UpdateRunRequest, JobRequest
from nbox.messages import rpc, read_file_to_binary, read_file_to_string, message_to_dict

class Tracer:
  def __init__(self, local: bool = False, start_heartbeat: bool = True, heartbeat_every: int = 60):
    self.heartbeat_every = heartbeat_every

    # create the kwargs that are used throughout
    self.job_proto = None
    self.run_id = None
    self.job_id = None
    self.workspace_id = None
    self.network_tracer = False
    self.trace_file = None

    # from the data available in the init folder, get the job proto
    if local:
      pass
    else:
      run_data = secret(AuthConfig.nbx_pod_run) # user should never have "run" on their local
      if run_data is not None:
        self.init(run_data, start_heartbeat)

  @classmethod
  def local(cls, job_proto: JobProto, run_id):
    tracer = Tracer(local = True)
    tracer.job_proto = job_proto
    tracer.run_id = run_id
    tracer.job_id = job_proto.id
    tracer.workspace_id = job_proto.auth_info.workspace_id
    tracer.local_init()
    return tracer

  def local_init(self):
    folder = U.join(f"{U.env.NBOX_HOME_DIR()}", "traces")
    os.makedirs(folder, exist_ok = True)
    file = U.join(folder, f"{self.run_id}.jsonl")

    logger.debug(f"Username: {self.job_proto.auth_info.username}")
    logger.debug(f"Job Id: {self.job_proto.id}")
    logger.debug(f"Run Id: {self.run_id}")
    logger.debug(f"Workspace Id: {self.job_proto.auth_info.workspace_id}")
    self.trace_file = open(file, "a")

  def init(self, run_data, start_heartbeat):
    init_folder = U.env.NBOX_JOB_FOLDER(".")
    if not init_folder:
      raise RuntimeError("NBOX_JOB_FOLDER not set")
    if not os.path.exists(init_folder):
      raise RuntimeError(f"NBOX_JOB_FOLDER {init_folder} does not exist")

    # get this data from the local secrets file
    self.job_id = run_data.get("job_id", None)
    self.run_id = run_data.get("token", None)

    # grandfather old messages (<v0.9.14rc13)
    fp_bin = U.join(init_folder, "job_proto.msg")
    fp_str = U.join(init_folder, "job_proto.pbtxt")
    if os.path.exists(fp_bin):
      self.job_proto: JobProto = read_file_to_binary(fp_bin, JobProto())
    elif os.path.exists(fp_str):
      self.job_proto: JobProto = read_file_to_string(fp_str, JobProto())
    else:
      logger.warning(f"Could not find job_proto.msg or job_proto.pbtxt in {init_folder}, fetching from database")
      self.job_proto = nbox_grpc_stub.GetJob(JobRequest(
        auth_info=auth_info_pb(),
        job = JobProto(id = self.job_id)
      ))

    self.job_proto.id = self.job_id # because when creating a new job, client does not know the ID
    self.workspace_id = secret(AuthConfig._workspace_id)
    self.network_tracer = True

    # logger.debug(f"Username: {self.job_proto.auth_info.username}")
    logger.info(f"Job Id (Run Id) [Workspace ID]: {self.job_id} ({self.run_id}) [{self.workspace_id}]")
    self.job_proto.status = JobProto.Status.ACTIVE # automatically first run will

    # start heartbeat in a different thread
    if start_heartbeat:
      self.thread = threading.Thread(target=self.hearbeat_thread_worker)
      self.thread.start()

  @property
  def active(self):
    return hasattr(self, "job_proto")

  def __repr__(self) -> str:
    return f"Tracer() for job {self.job_id}"

  def _rpc(self, message: str = ""):
    try:
      rpc(
        nbox_grpc_stub.UpdateRun,
        UpdateRunRequest(
          token = self.run_id,
          job = self.job_proto,
          updated_at = SimplerTimes.get_now_pb(),
          auth_info = auth_info_pb()
        ),
        message or f"Could not update job {self.job_proto.id}",
        raise_on_error = True
      )
    except Exception as e:
      logger.error(f"Could not update job {self.job_proto.id}\n  Error: {e}\n  Most likely could not commumicate with the server")
      U.hard_exit_program(1)

  def __call__(self, node: Node, verbose: bool = False):
    if self.network_tracer:
      self.job_proto.dag.flowchart.nodes[node.id].CopyFrom(node) # even if fails we can keep caching this
      self._rpc()
    else:
      self.trace_file.write(dumps(message_to_dict(node)) + "\n")

  def hearbeat_thread_worker(self):
    while True:
      self._rpc()
      sleep(self.heartbeat_every)

  def stop(self):
    if not self.network_tracer:
      return
    self.thread.join()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ..common import common_pb2 as common_dot_common__pb2
from ..deploy import serve_pb2 as deploy_dot_serve__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


class ServingServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Create = channel.unary_unary(
                '/deploy.ServingService/Create',
                request_serializer=deploy_dot_serve__pb2.ServingRequest.SerializeToString,
                response_deserializer=deploy_dot_serve__pb2.Serving.FromString,
                )
        self.GetServing = channel.unary_unary(
                '/deploy.ServingService/GetServing',
                request_serializer=deploy_dot_serve__pb2.ServingRequest.SerializeToString,
                response_deserializer=deploy_dot_serve__pb2.Serving.FromString,
                )
        self.ListServings = channel.unary_unary(
                '/deploy.ServingService/ListServings',
                request_serializer=deploy_dot_serve__pb2.ServingListRequest.SerializeToString,
                response_deserializer=deploy_dot_serve__pb2.ServingListResponse.FromString,
                )
        self.DeleteServing = channel.unary_unary(
                '/deploy.ServingService/DeleteServing',
                request_serializer=deploy_dot_serve__pb2.ServingRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.UpdateServing = channel.unary_unary(
                '/deploy.ServingService/UpdateServing',
                request_serializer=deploy_dot_serve__pb2.UpdateServingRequest.SerializeToString,
                response_deserializer=deploy_dot_serve__pb2.Serving.FromString,
                )


class ServingServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Create(self, request, context):
        """Creates a new Serving Group.
        If `.models` are specified, the Serving Group will be created with new
        `len(.models)` models.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetServing(self, request, context):
        """Get all available info for a Serving Group. All `.models` are populated.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListServings(self, request, context):
        """Get list of Serving Grpup(s) of a user/workspace.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteServing(self, request, context):
        """Delete a Serving Group and all its models.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateServing(self, request, context):
        """Update a Serving Group. Only the fields specified in the `update_mask` are
        updated.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ServingServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Create': grpc.unary_unary_rpc_method_handler(
                    servicer.Create,
                    request_deserializer=deploy_dot_serve__pb2.ServingRequest.FromString,
                    response_serializer=deploy_dot_serve__pb2.Serving.SerializeToString,
            ),
            'GetServing': grpc.unary_unary_rpc_method_handler(
                    servicer.GetServing,
                    request_deserializer=deploy_dot_serve__pb2.ServingRequest.FromString,
                    response_serializer=deploy_dot_serve__pb2.Serving.SerializeToString,
            ),
            'ListServings': grpc.unary_unary_rpc_method_handler(
                    servicer.ListServings,
                    request_deserializer=deploy_dot_serve__pb2.ServingListRequest.FromString,
                    response_serializer=deploy_dot_serve__pb2.ServingListResponse.SerializeToString,
            ),
            'DeleteServing': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteServing,
                    request_deserializer=deploy_dot_serve__pb2.ServingRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'UpdateServing': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateServing,
                    request_deserializer=deploy_dot_serve__pb2.UpdateServingRequest.FromString,
                    response_serializer=deploy_dot_serve__pb2.Serving.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'deploy.ServingService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ServingService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Create(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/deploy.ServingService/Create',
            deploy_dot_serve__pb2.ServingRequest.SerializeToString,
            deploy_dot_serve__pb2.Serving.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetServing(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/deploy.ServingService/GetServing',
            deploy_dot_serve__pb2.ServingRequest.SerializeToString,
            deploy_dot_serve__pb2.Serving.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListServings(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/deploy.ServingService/ListServings',
            deploy_dot_serve__pb2.ServingListRequest.SerializeToString,
            deploy_dot_serve__pb2.ServingListResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteServing(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/deploy.ServingService/DeleteServing',
            deploy_dot_serve__pb2.ServingRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateServing(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/deploy.ServingService/UpdateServing',
            deploy_dot_serve__pb2.UpdateServingRequest.SerializeToString,
            deploy_dot_serve__pb2.Serving.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)


class ModelServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetModel = channel.unary_unary(
                '/deploy.ModelService/GetModel',
                request_serializer=deploy_dot_serve__pb2.ModelRequest.SerializeToString,
                response_deserializer=deploy_dot_serve__pb2.Model.FromString,
                )
        self.Watch = channel.unary_stream(
                '/deploy.ModelService/Watch',
                request_serializer=deploy_dot_serve__pb2.ModelRequest.SerializeToString,
                response_deserializer=deploy_dot_serve__pb2.ServingStatus.FromString,
                )
        self.UploadModel = channel.unary_unary(
                '/deploy.ModelService/UploadModel',
                request_serializer=deploy_dot_serve__pb2.ModelRequest.SerializeToString,
                response_deserializer=deploy_dot_serve__pb2.Model.FromString,
                )
        self.Deploy = channel.unary_unary(
                '/deploy.ModelService/Deploy',
                request_serializer=deploy_dot_serve__pb2.ModelRequest.SerializeToString,
                response_deserializer=deploy_dot_serve__pb2.Model.FromString,
                )
        self.ModelLogs = channel.unary_stream(
                '/deploy.ModelService/ModelLogs',
                request_serializer=deploy_dot_serve__pb2.ModelRequest.SerializeToString,
                response_deserializer=common_dot_common__pb2.Log.FromString,
                )
        self.ListModels = channel.unary_unary(
                '/deploy.ModelService/ListModels',
                request_serializer=deploy_dot_serve__pb2.ModelListRequest.SerializeToString,
                response_deserializer=deploy_dot_serve__pb2.ModelListResponse.FromString,
                )
        self.DeleteModel = channel.unary_unary(
                '/deploy.ModelService/DeleteModel',
                request_serializer=deploy_dot_serve__pb2.ModelRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.SetModelPin = channel.unary_unary(
                '/deploy.ModelService/SetModelPin',
                request_serializer=deploy_dot_serve__pb2.ModelRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.UpdateModel = channel.unary_unary(
                '/deploy.ModelService/UpdateModel',
                request_serializer=deploy_dot_serve__pb2.UpdateModelRequest.SerializeToString,
                response_deserializer=deploy_dot_serve__pb2.Model.FromString,
                )


class ModelServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetModel(self, request, context):
        """Get all available info for a model.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Watch(self, request, context):
        """Watch for status update of a model in Serving Group.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UploadModel(self, request, context):
        """Get S3 Upload URL for new/existing Model in Serving Group.
        If id is none then creates new Model with id `id` in the Serving Group.
        If model's serving_group_id is none/doesn't exist, then throws error.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Deploy(self, request, context):
        """Deploy new model in Serving Group.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ModelLogs(self, request, context):
        """Get logs of a model.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListModels(self, request, context):
        """Get list of models in a Serving Group.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteModel(self, request, context):
        """Delete a model in a Serving Group.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetModelPin(self, request, context):
        """Pinned a model of a Serving Group.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateModel(self, request, context):
        """Update a model of a Serving Group.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ModelServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetModel': grpc.unary_unary_rpc_method_handler(
                    servicer.GetModel,
                    request_deserializer=deploy_dot_serve__pb2.ModelRequest.FromString,
                    response_serializer=deploy_dot_serve__pb2.Model.SerializeToString,
            ),
            'Watch': grpc.unary_stream_rpc_method_handler(
                    servicer.Watch,
                    request_deserializer=deploy_dot_serve__pb2.ModelRequest.FromString,
                    response_serializer=deploy_dot_serve__pb2.ServingStatus.SerializeToString,
            ),
            'UploadModel': grpc.unary_unary_rpc_method_handler(
                    servicer.UploadModel,
                    request_deserializer=deploy_dot_serve__pb2.ModelRequest.FromString,
                    response_serializer=deploy_dot_serve__pb2.Model.SerializeToString,
            ),
            'Deploy': grpc.unary_unary_rpc_method_handler(
                    servicer.Deploy,
                    request_deserializer=deploy_dot_serve__pb2.ModelRequest.FromString,
                    response_serializer=deploy_dot_serve__pb2.Model.SerializeToString,
            ),
            'ModelLogs': grpc.unary_stream_rpc_method_handler(
                    servicer.ModelLogs,
                    request_deserializer=deploy_dot_serve__pb2.ModelRequest.FromString,
                    response_serializer=common_dot_common__pb2.Log.SerializeToString,
            ),
            'ListModels': grpc.unary_unary_rpc_method_handler(
                    servicer.ListModels,
                    request_deserializer=deploy_dot_serve__pb2.ModelListRequest.FromString,
                    response_serializer=deploy_dot_serve__pb2.ModelListResponse.SerializeToString,
            ),
            'DeleteModel': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteModel,
                    request_deserializer=deploy_dot_serve__pb2.ModelRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'SetModelPin': grpc.unary_unary_rpc_method_handler(
                    servicer.SetModelPin,
                    request_deserializer=deploy_dot_serve__pb2.ModelRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'UpdateModel': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateModel,
                    request_deserializer=deploy_dot_serve__pb2.UpdateModelRequest.FromString,
                    response_serializer=deploy_dot_serve__pb2.Model.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'deploy.ModelService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ModelService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/deploy.ModelService/GetModel',
            deploy_dot_serve__pb2.ModelRequest.SerializeToString,
            deploy_dot_serve__pb2.Model.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Watch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/deploy.ModelService/Watch',
            deploy_dot_serve__pb2.ModelRequest.SerializeToString,
            deploy_dot_serve__pb2.ServingStatus.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UploadModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/deploy.ModelService/UploadModel',
            deploy_dot_serve__pb2.ModelRequest.SerializeToString,
            deploy_dot_serve__pb2.Model.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Deploy(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/deploy.ModelService/Deploy',
            deploy_dot_serve__pb2.ModelRequest.SerializeToString,
            deploy_dot_serve__pb2.Model.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ModelLogs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/deploy.ModelService/ModelLogs',
            deploy_dot_serve__pb2.ModelRequest.SerializeToString,
            common_dot_common__pb2.Log.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListModels(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/deploy.ModelService/ListModels',
            deploy_dot_serve__pb2.ModelListRequest.SerializeToString,
            deploy_dot_serve__pb2.ModelListResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/deploy.ModelService/DeleteModel',
            deploy_dot_serve__pb2.ModelRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetModelPin(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/deploy.ModelService/SetModelPin',
            deploy_dot_serve__pb2.ModelRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/deploy.ModelService/UpdateModel',
            deploy_dot_serve__pb2.UpdateModelRequest.SerializeToString,
            deploy_dot_serve__pb2.Model.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



"""
This file contains the code for all the compatibility with the [huggingface platform](https://huggingface.co/)
"""

try:
  import huggingface_hub as hfhub
  from transformers.trainer_callback import TrainerCallback, TrainerControl, TrainerState
  from transformers.training_args import TrainingArguments
except ImportError:
  raise ImportError("Please install the huggingface_hub and transformers packages to use this integration")

from nbox.utils import logger
from nbox.projects import Project


class _Cfg:
  debug: bool = False
  on_init_end: bool = True
  on_train_begin: bool = True
  on_train_end: bool = True
  on_epoch_begin: bool = False
  on_epoch_end: bool = False
  on_step_begin: bool = False
  on_substep_end: bool = False
  on_step_end: bool = False
  on_evaluate: bool = True
  on_predict: bool = False
  on_save: bool = True
  on_log: bool = True
  on_prediction_step: bool = False
  def __init__(self, **kwargs):
    for k, v in kwargs.items():
      setattr(self, k, v)

class NimbleBoxCallback(TrainerCallback):
  """
  This `TrainerCallback` talks to the [NimbleBox server](https://nimblebox.ai) to log training metrics and checkpoints.

  This will automatically detect if you are running on NimbleBox pods and will automatically pick the appropriate
  values from the environment variables such as your project ID, Run ID, etc. If you are running this outside of the
  NimbleBox, you can read more in our [docs](https://nimblebox.ai/docs).
  """

  def __init__(
    self,
    on_init_end: bool = True,
    on_train_begin: bool = True,
    on_train_end: bool = True,
    on_epoch_begin: bool = False,
    on_epoch_end: bool = False,
    on_step_begin: bool = False,
    on_substep_end: bool = False,
    on_step_end: bool = False,
    on_evaluate: bool = True,
    on_predict: bool = False,
    on_save: bool = True,
    on_log: bool = True,
    on_prediction_step: bool = False,
    *,
    debug: bool = False,
  ) -> None:
    self._initialized = False
    self.cfg: _Cfg(
      debug = debug,
      on_init_end = on_init_end,
      on_train_begin = on_train_begin,
      on_train_end = on_train_end,
      on_epoch_begin = on_epoch_begin,
      on_epoch_end = on_epoch_end,
      on_step_begin = on_step_begin,
      on_substep_end = on_substep_end,
      on_step_end = on_step_end,
      on_evaluate = on_evaluate,
      on_predict = on_predict,
      on_save = on_save,
      on_log = on_log,
      on_prediction_step = on_prediction_step,
    )

  def setup(self, args, state, model, tokenizer, **kwargs):
    if self._initialized:
      return
    if state.is_world_process_zero:
      self.p = Project()

  def on_init_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
    if not (state.is_world_process_zero or self.cfg.on_init_end):
      return
    logger.info("on_init_end: initialization of the Trainer is complete. Will load all the training arguments.")
    metadata = {
      "hf_training_args": args.to_dict()
    }
    self._tracker = self.p.get_exp_tracker(metadata = metadata)

  def on_train_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
    if not (state.is_world_process_zero or self.cfg.on_train_begin):
      return
    logger.debug("on_train_begin: training is about to begin.")

  def on_train_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
    if not (state.is_world_process_zero or self.cfg.on_train_end):
      return
    logger.debug("on_train_end: training is complete.")

  def on_epoch_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
    if not (state.is_world_process_zero or self.cfg.on_epoch_begin):
      return
    logger.debug("on_epoch_begin: epoch is about to begin.")

  def on_epoch_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
    if not (state.is_world_process_zero or self.cfg.on_epoch_end):
      return
    logger.debug("on_epoch_end: epoch is complete.")

  def on_step_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
    if not (state.is_world_process_zero or self.cfg.on_step_begin):
      return
    logger.debug("on_step_begin: step is about to begin.")

  def on_substep_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
    if not (state.is_world_process_zero or self.cfg.on_substep_end):
      return
    logger.debug("on_substep_end: substep is complete.")

  def on_step_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
    if not (state.is_world_process_zero or self.cfg.on_step_end):
      return
    logger.debug("on_step_end: step is complete.")

  def on_evaluate(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
    if not (state.is_world_process_zero or self.cfg.on_evaluate):
      return
    logger.debug("on_evaluate: evaluation is complete.")

  def on_predict(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, metrics, **kwargs):
    if not (state.is_world_process_zero or self.cfg.on_predict):
      return
    logger.debug("on_predict: prediction is complete.")

  def on_save(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
    if not (state.is_world_process_zero or self.cfg.on_save):
      return
    logger.debug("on_save: checkpoint is saved.")

  def on_log(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, model=None, tokenizer=None, logs=None, **kwargs):
    if not (state.is_world_process_zero or self.cfg.on_log):
      return
    logger.debug("on_log: logging is complete.")
    self._tracker.log(logs)

  def on_prediction_step(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
    if not (state.is_world_process_zero or self.cfg.on_prediction_step):
      return
    logger.debug("on_prediction_step")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/common.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12proto/common.proto\"/\n\x0b\x41\x63knowledge\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\tb\x06proto3')



_ACKNOWLEDGE = DESCRIPTOR.message_types_by_name['Acknowledge']
Acknowledge = _reflection.GeneratedProtocolMessageType('Acknowledge', (_message.Message,), {
  'DESCRIPTOR' : _ACKNOWLEDGE,
  '__module__' : 'proto.common_pb2'
  # @@protoc_insertion_point(class_scope:Acknowledge)
  })
_sym_db.RegisterMessage(Acknowledge)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ACKNOWLEDGE._serialized_start=22
  _ACKNOWLEDGE._serialized_end=69
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: lmao_v2.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rlmao_v2.proto\x12\x07lmao_v2\x1a\x1fgoogle/protobuf/timestamp.proto\"\x83\x02\n\x07Project\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x14\n\x0cproject_name\x18\x03 \x01(\t\x12\x12\n\ntotal_runs\x18\x04 \x01(\x05\x12\x16\n\x0etotal_servings\x18\t \x01(\x05\x12\x13\n\x0blog_metrics\x18\x05 \x03(\t\x12\x16\n\x0e\x63onfig_metrics\x18\x06 \x03(\t\x12\x17\n\x0fserving_metrics\x18\n \x03(\t\x12\x0c\n\x04tags\x18\x0b \x03(\t\x12\x12\n\ncreated_at\x18\x07 \x01(\x03\x12\x12\n\nis_deleted\x18\x08 \x01(\x08\x12\x10\n\x08relic_id\x18\x0c \x01(\t\"*\n\x12InitProjectRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\"e\n\x13InitProjectResponse\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x10\n\x08relic_id\x18\x03 \x01(\t\x12\x12\n\nrelic_name\x18\x04 \x01(\t\"X\n\x13ListProjectsRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x1a\n\x12project_id_or_name\x18\x02 \x01(\t\x12\x0f\n\x07page_no\x18\x03 \x01(\x05\"O\n\x14ListProjectsResponse\x12\"\n\x08projects\x18\x01 \x03(\x0b\x32\x10.lmao_v2.Project\x12\x13\n\x0btotal_pages\x18\x02 \x01(\x05\"\xda\x01\n\x0c\x41gentDetails\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\'\n\x04type\x18\x02 \x01(\x0e\x32\x19.lmao_v2.AgentDetails.NBX\x12\x14\n\nnbx_job_id\x18\x65 \x01(\tH\x00\x12\x18\n\x0enbx_serving_id\x18\x66 \x01(\tH\x00\x12\x14\n\nnbx_run_id\x18g \x01(\tH\x01\x12\x16\n\x0cnbx_model_id\x18h \x01(\tH\x01\"\x1b\n\x03NBX\x12\x07\n\x03JOB\x10\x00\x12\x0b\n\x07SERVING\x10\x01\x42\x04\n\x02idB\n\n\x08instance\"/\n\x0b\x41\x63knowledge\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\"\xb5\x02\n\x0cRecordColumn\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x32\n\nvalue_type\x18\x02 \x01(\x0e\x32\x1e.lmao_v2.RecordColumn.DataType\x12-\n\x04rows\x18\x03 \x03(\x0b\x32\x1f.lmao_v2.RecordColumn.RecordRow\x1a\x84\x01\n\tRecordRow\x12\t\n\x01x\x18\x01 \x01(\x03\x12-\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\nfloat_data\x18\x64 \x01(\x02\x12\x14\n\x0cinteger_data\x18\x65 \x01(\x05\x12\x13\n\x0bstring_data\x18\x66 \x01(\t\".\n\x08\x44\x61taType\x12\t\n\x05\x46LOAT\x10\x00\x12\x0b\n\x07INTEGER\x10\x01\x12\n\n\x06STRING\x10\x02\"\xef\x01\n\x06Record\x12\x0b\n\x03key\x18\x01 \x01(\t\x12,\n\nvalue_type\x18\x02 \x01(\x0e\x32\x18.lmao_v2.Record.DataType\x12\x0c\n\x04step\x18\x03 \x01(\x03\x12-\n\ttimestamp\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\nfloat_data\x18\x64 \x03(\x02\x12\x14\n\x0cinteger_data\x18\x65 \x03(\x05\x12\x13\n\x0bstring_data\x18\x66 \x03(\t\".\n\x08\x44\x61taType\x12\t\n\x05\x46LOAT\x10\x00\x12\x0b\n\x07INTEGER\x10\x01\x12\n\n\x06STRING\x10\x02\"\xfd\x01\n\x06RunLog\x12\x14\n\x0cworkspace_id\x18\x06 \x01(\t\x12\x12\n\nproject_id\x18\x05 \x01(\t\x12\x15\n\rexperiment_id\x18\x01 \x01(\t\x12\x12\n\nserving_id\x18\x07 \x01(\t\x12\x1d\n\x04\x64\x61ta\x18\x02 \x03(\x0b\x32\x0f.lmao_v2.Record\x12*\n\x0b\x63olumn_data\x18\x03 \x03(\x0b\x32\x15.lmao_v2.RecordColumn\x12)\n\x08log_type\x18\x04 \x01(\x0e\x32\x17.lmao_v2.RunLog.LogType\"(\n\x07LogType\x12\n\n\x06SYSTEM\x10\x00\x12\x07\n\x03NBX\x10\x01\x12\x08\n\x04USER\x10\x02\"\xdf\x02\n\x07Serving\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x14\n\x0cproject_name\x18\x03 \x01(\t\x12\x12\n\nserving_id\x18\x04 \x01(\t\x12$\n\x05\x61gent\x18\x05 \x01(\x0b\x32\x15.lmao_v2.AgentDetails\x12\x12\n\ncreated_at\x18\x06 \x01(\x03\x12\x0f\n\x07updated\x18\x07 \x01(\x03\x12\x10\n\x08\x65nded_at\x18\x08 \x01(\x03\x12\x15\n\rsave_location\x18\t \x01(\t\x12\x0e\n\x06\x63onfig\x18\n \x01(\t\x12\'\n\x06status\x18\x0b \x01(\x0e\x32\x17.lmao_v2.Serving.Status\x12\x13\n\x0bupdate_keys\x18\x64 \x03(\t\">\n\x06Status\x12\x0b\n\x07NOT_SET\x10\x00\x12\x0b\n\x07RUNNING\x10\x01\x12\r\n\tCOMPLETED\x10\x02\x12\x0b\n\x07\x44\x45LETED\x10\x03\"^\n\x13ListServingsRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x0f\n\x07page_no\x18\x03 \x01(\x05\x12\x0c\n\x04\x64\x65sc\x18\x04 \x01(\x08\"b\n\x14ListServingsResponse\x12\x13\n\x0bserving_ids\x18\x01 \x03(\t\x12\x0c\n\x04keys\x18\x02 \x03(\t\x12\x12\n\ncreated_at\x18\x03 \x03(\x03\x12\x13\n\x0btotal_pages\x18\x04 \x01(\x05\"\xc4\x01\n\x11ServingLogRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x12\n\nserving_id\x18\x03 \x01(\t\x12\x0b\n\x03key\x18\x04 \x01(\t\x12)\n\x05\x61\x66ter\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12*\n\x06\x62\x65\x66ore\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\r\n\x05limit\x18\x07 \x01(\x05\"3\n\x12ServingLogResponse\x12\x1d\n\x04logs\x18\x02 \x03(\x0b\x32\x0f.lmao_v2.RunLog\"\xef\x02\n\x03Run\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x14\n\x0cproject_name\x18\x03 \x01(\t\x12\x15\n\rexperiment_id\x18\x04 \x01(\t\x12$\n\x05\x61gent\x18\x05 \x01(\x0b\x32\x15.lmao_v2.AgentDetails\x12\x12\n\ncreated_at\x18\x06 \x01(\x03\x12\x12\n\nupdated_at\x18\x07 \x01(\x03\x12\x10\n\x08\x65nded_at\x18\x08 \x01(\x03\x12\x15\n\rsave_location\x18\t \x01(\t\x12\x0e\n\x06\x63onfig\x18\x0b \x01(\t\x12#\n\x06status\x18\x0c \x01(\x0e\x32\x13.lmao_v2.Run.Status\x12\x13\n\x0bupdate_keys\x18\x64 \x03(\t\"J\n\x06Status\x12\x0b\n\x07NOT_SET\x10\x00\x12\x0b\n\x07RUNNING\x10\x01\x12\r\n\tCOMPLETED\x10\x02\x12\n\n\x06\x46\x41ILED\x10\x03\x12\x0b\n\x07\x44\x45LETED\x10\x04J\x04\x08\n\x10\x0b\"\x8e\x01\n\x0eInitRunRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x14\n\x0cproject_name\x18\x03 \x01(\t\x12,\n\ragent_details\x18\x04 \x01(\x0b\x32\x15.lmao_v2.AgentDetails\x12\x0e\n\x06\x63onfig\x18\x06 \x01(\t\"q\n\x0fListRunsRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x15\n\rexperiment_id\x18\x03 \x01(\t\x12\x0f\n\x07page_no\x18\x04 \x01(\x05\x12\x0c\n\x04\x64\x65sc\x18\x05 \x01(\x08\"a\n\x10ListRunsResponse\x12\x16\n\x0e\x65xperiment_ids\x18\x01 \x03(\t\x12\x0c\n\x04keys\x18\x02 \x03(\t\x12\x12\n\ncreated_at\x18\x03 \x03(\x03\x12\x13\n\x0btotal_pages\x18\x04 \x01(\x05\"\x7f\n\rRunLogRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x15\n\rexperiment_id\x18\x03 \x01(\t\x12\x0b\n\x03key\x18\x04 \x01(\t\x12\x10\n\x08start_at\x18\x05 \x01(\x05\x12\x0e\n\x06\x65nd_at\x18\x06 \x01(\x05\"\x89\x01\n\x19GetExperimentTableRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x0f\n\x07page_no\x18\x03 \x01(\x05\x12\x15\n\rexperiment_id\x18\x06 \x01(\t\x12\r\n\x05start\x18\x04 \x01(\x05\x12\x0b\n\x03\x65nd\x18\x05 \x01(\x05\"\x95\x01\n\x0f\x45xperimentTable\x12\x13\n\x0btotal_pages\x18\x01 \x01(\x05\x12*\n\x04rows\x18\x02 \x03(\x0b\x32\x1c.lmao_v2.ExperimentTable.Row\x1a\x41\n\x03Row\x12\x15\n\rexperiment_id\x18\x01 \x01(\t\x12#\n\x04\x64\x61ta\x18\x02 \x03(\x0b\x32\x15.lmao_v2.RecordColumn\"\x8f\x02\n\x04Rule\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x11\n\trule_name\x18\x03 \x01(\t\x12\x12\n\ncreated_at\x18\x04 \x01(\x03\x12\x16\n\x0eis_deactivated\x18\x05 \x01(\x08\x12\x12\n\nis_deleted\x18\x06 \x01(\x08\x12\r\n\x05logic\x18\x07 \x01(\t\x12\x11\n\temail_ids\x18\x08 \x03(\t\x12\x15\n\ralert_message\x18\t \x01(\t\x12\x12\n\ntrue_count\x18\n \x01(\x05\x12\x15\n\rlogic_version\x18\x0b \x01(\x05\x12\x11\n\tmax_range\x18\x0c \x01(\x05\x12\x13\n\x0bupdate_keys\x18\x64 \x03(\t\"\x8c\x01\n\tListRules\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x0f\n\x07page_no\x18\x03 \x01(\x05\x12\x13\n\x0btotal_pages\x18\x04 \x01(\x05\x12\x1c\n\x05rules\x18\x05 \x03(\x0b\x32\r.lmao_v2.Rule\x12\x11\n\trule_name\x18\x06 \x01(\t\"\xf9\x01\n\x0bRuleBuilder\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nproject_id\x18\x02 \x01(\t\x12\x0f\n\x07metrics\x18\x03 \x03(\t\x12*\n\toperators\x18\x04 \x03(\x0b\x32\x17.lmao_v2.RuleBuilder.Op\x12*\n\tfunctions\x18\x05 \x03(\x0b\x32\x17.lmao_v2.RuleBuilder.Fn\x1a\"\n\x02Op\x12\x0e\n\x06symbol\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x1a\x33\n\x02\x46n\x12\x0e\n\x06symbol\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07\x66\x65_meta\x18\x03 \x01(\t2\xd2\x0c\n\x04LMAO\x12K\n\x0cinit_project\x12\x1b.lmao_v2.InitProjectRequest\x1a\x1c.lmao_v2.InitProjectResponse\"\x00\x12\x33\n\x0bget_project\x12\x10.lmao_v2.Project\x1a\x10.lmao_v2.Project\"\x00\x12N\n\rlist_projects\x12\x1c.lmao_v2.ListProjectsRequest\x1a\x1d.lmao_v2.ListProjectsResponse\"\x00\x12:\n\x0e\x64\x65lete_project\x12\x10.lmao_v2.Project\x1a\x14.lmao_v2.Acknowledge\"\x00\x12\x33\n\x08init_run\x12\x17.lmao_v2.InitRunRequest\x1a\x0c.lmao_v2.Run\"\x00\x12\x39\n\x11update_run_status\x12\x0c.lmao_v2.Run\x1a\x14.lmao_v2.Acknowledge\"\x00\x12\x31\n\x06on_log\x12\x0f.lmao_v2.RunLog\x1a\x14.lmao_v2.Acknowledge\"\x00\x12\x34\n\x0con_train_end\x12\x0c.lmao_v2.Run\x1a\x14.lmao_v2.Acknowledge\"\x00\x12\x42\n\tlist_runs\x12\x18.lmao_v2.ListRunsRequest\x1a\x19.lmao_v2.ListRunsResponse\"\x00\x12/\n\x0fget_run_details\x12\x0c.lmao_v2.Run\x1a\x0c.lmao_v2.Run\"\x00\x12\x38\n\x0bget_run_log\x12\x16.lmao_v2.RunLogRequest\x1a\x0f.lmao_v2.RunLog\"\x00\x12V\n\x14get_experiment_table\x12\".lmao_v2.GetExperimentTableRequest\x1a\x18.lmao_v2.ExperimentTable\"\x00\x12\x39\n\x11\x64\x65lete_experiment\x12\x0c.lmao_v2.Run\x1a\x14.lmao_v2.Acknowledge\"\x00\x12;\n\x0cinit_serving\x12\x17.lmao_v2.InitRunRequest\x1a\x10.lmao_v2.Serving\"\x00\x12\x41\n\x15update_serving_status\x12\x10.lmao_v2.Serving\x1a\x14.lmao_v2.Acknowledge\"\x00\x12\x39\n\x0eon_serving_log\x12\x0f.lmao_v2.RunLog\x1a\x14.lmao_v2.Acknowledge\"\x00\x12:\n\x0eon_serving_end\x12\x10.lmao_v2.Serving\x1a\x14.lmao_v2.Acknowledge\"\x00\x12N\n\rlist_servings\x12\x1c.lmao_v2.ListServingsRequest\x1a\x1d.lmao_v2.ListServingsResponse\"\x00\x12;\n\x13get_serving_details\x12\x10.lmao_v2.Serving\x1a\x10.lmao_v2.Serving\"\x00\x12L\n\x0fget_serving_log\x12\x1a.lmao_v2.ServingLogRequest\x1a\x1b.lmao_v2.ServingLogResponse\"\x00\x12:\n\x0e\x64\x65lete_serving\x12\x10.lmao_v2.Serving\x1a\x14.lmao_v2.Acknowledge\"\x00\x12@\n\x10get_rule_builder\x12\x14.lmao_v2.RuleBuilder\x1a\x14.lmao_v2.RuleBuilder\"\x00\x12-\n\x0b\x63reate_rule\x12\r.lmao_v2.Rule\x1a\r.lmao_v2.Rule\"\x00\x12\x36\n\nlist_rules\x12\x12.lmao_v2.ListRules\x1a\x12.lmao_v2.ListRules\"\x00\x12\x34\n\x0b\x64\x65lete_rule\x12\r.lmao_v2.Rule\x1a\x14.lmao_v2.Acknowledge\"\x00\x12\x34\n\x0bupdate_rule\x12\r.lmao_v2.Rule\x1a\x14.lmao_v2.Acknowledge\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'lmao_v2_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _PROJECT._serialized_start=60
  _PROJECT._serialized_end=319
  _INITPROJECTREQUEST._serialized_start=321
  _INITPROJECTREQUEST._serialized_end=363
  _INITPROJECTRESPONSE._serialized_start=365
  _INITPROJECTRESPONSE._serialized_end=466
  _LISTPROJECTSREQUEST._serialized_start=468
  _LISTPROJECTSREQUEST._serialized_end=556
  _LISTPROJECTSRESPONSE._serialized_start=558
  _LISTPROJECTSRESPONSE._serialized_end=637
  _AGENTDETAILS._serialized_start=640
  _AGENTDETAILS._serialized_end=858
  _AGENTDETAILS_NBX._serialized_start=813
  _AGENTDETAILS_NBX._serialized_end=840
  _ACKNOWLEDGE._serialized_start=860
  _ACKNOWLEDGE._serialized_end=907
  _RECORDCOLUMN._serialized_start=910
  _RECORDCOLUMN._serialized_end=1219
  _RECORDCOLUMN_RECORDROW._serialized_start=1039
  _RECORDCOLUMN_RECORDROW._serialized_end=1171
  _RECORDCOLUMN_DATATYPE._serialized_start=1173
  _RECORDCOLUMN_DATATYPE._serialized_end=1219
  _RECORD._serialized_start=1222
  _RECORD._serialized_end=1461
  _RECORD_DATATYPE._serialized_start=1173
  _RECORD_DATATYPE._serialized_end=1219
  _RUNLOG._serialized_start=1464
  _RUNLOG._serialized_end=1717
  _RUNLOG_LOGTYPE._serialized_start=1677
  _RUNLOG_LOGTYPE._serialized_end=1717
  _SERVING._serialized_start=1720
  _SERVING._serialized_end=2071
  _SERVING_STATUS._serialized_start=2009
  _SERVING_STATUS._serialized_end=2071
  _LISTSERVINGSREQUEST._serialized_start=2073
  _LISTSERVINGSREQUEST._serialized_end=2167
  _LISTSERVINGSRESPONSE._serialized_start=2169
  _LISTSERVINGSRESPONSE._serialized_end=2267
  _SERVINGLOGREQUEST._serialized_start=2270
  _SERVINGLOGREQUEST._serialized_end=2466
  _SERVINGLOGRESPONSE._serialized_start=2468
  _SERVINGLOGRESPONSE._serialized_end=2519
  _RUN._serialized_start=2522
  _RUN._serialized_end=2889
  _RUN_STATUS._serialized_start=2809
  _RUN_STATUS._serialized_end=2883
  _INITRUNREQUEST._serialized_start=2892
  _INITRUNREQUEST._serialized_end=3034
  _LISTRUNSREQUEST._serialized_start=3036
  _LISTRUNSREQUEST._serialized_end=3149
  _LISTRUNSRESPONSE._serialized_start=3151
  _LISTRUNSRESPONSE._serialized_end=3248
  _RUNLOGREQUEST._serialized_start=3250
  _RUNLOGREQUEST._serialized_end=3377
  _GETEXPERIMENTTABLEREQUEST._serialized_start=3380
  _GETEXPERIMENTTABLEREQUEST._serialized_end=3517
  _EXPERIMENTTABLE._serialized_start=3520
  _EXPERIMENTTABLE._serialized_end=3669
  _EXPERIMENTTABLE_ROW._serialized_start=3604
  _EXPERIMENTTABLE_ROW._serialized_end=3669
  _RULE._serialized_start=3672
  _RULE._serialized_end=3943
  _LISTRULES._serialized_start=3946
  _LISTRULES._serialized_end=4086
  _RULEBUILDER._serialized_start=4089
  _RULEBUILDER._serialized_end=4338
  _RULEBUILDER_OP._serialized_start=4251
  _RULEBUILDER_OP._serialized_end=4285
  _RULEBUILDER_FN._serialized_start=4287
  _RULEBUILDER_FN._serialized_end=4338
  _LMAO._serialized_start=4341
  _LMAO._serialized_end=5959
# @@protoc_insertion_point(module_scope)



from nbox.lmao.exp import Lmao, _lmaoConfig
from nbox.lmao.live import LmaoLive
from nbox.lmao.cli import LmaoCLI
from nbox.lmao.common import get_lmao_stub, get_git_details, get_record, ExperimentConfig, LMAO_RM_PREFIX
from nbox.sublime.proto import lmao_v2_pb2


"""
### Shell

This module contains the Operators that can be used to interact with the shell.
"""
import sys
from subprocess import Popen

from nbox import Operator
from nbox.operator import operator

class ShellCommand(Operator):
  def __init__(self, *commands):
    """Run multiple shell commands, uses `shelex` to prevent injection"""
    super().__init__()
    import string

    self.commands = commands
    all_in = []
    for c in self.commands:
      all_in.extend([tup[1] for tup in string.Formatter().parse(c) if tup[1] is not None])

  def forward(self, *args, **kwargs):
    import shlex
    import subprocess

    for comm in self.commands:
      comm = comm.format(*args, **kwargs)
      comm = shlex.split(comm)
      called_process: subprocess.CompletedProcess = subprocess.run(comm, check = True)
      if called_process.returncode != 0:
        raise Exception(f"Command {comm} failed with return code {called_process.stdout}")

class PythonScript(Operator):
  def __init__(self, fpath, python_path = "/usr/bin/python3", **kwargs):
    """Run any file in python as an operator"""
    super().__init__()
    self.fp = fpath
    self.python_path = python_path
    self.kwargs = kwargs

    self.kwargs_strings = []
    for k,v in kwargs.items():
      if v != None:
        self.kwargs_strings.append(f"--{k}={v}")
      else:
        self.kwargs_strings.append(f"--{k}")
    kwargs = "\ \n".join(self.kwargs_strings)
    self.com = ShellCommand(f"{python_path} {fpath} {kwargs}")

  def forward(self):
    return self.com()

class NBToScript(Operator):
  def __init__(self, notebook_path: str, py_file: str) -> None:
    super().__init__()
    self.nb = notebook_path
    self.trg = py_file

  def forward(self):
    import json

    with open(self.nb, "r") as f:
      data = json.load(f)
    
    code_script = "\n# This is an Autogenerated file by NBToScript Operators\n# Find out more: nimblebox.ai\n"
    # here we can start iterating over the code cells and fill up a script
    for cell in data["cells"]:
      if cell["cell_type"] == "code":
        code_script += "".join(cell["source"]) + "\n"

    with open(self.trg, "w") as f:
      f.write(code_script)

@operator()
def git_clone(
  repository: str,
  path: str = ".",
  branch: str = "",
  depth: int = 1,
  **kwargs
) -> bool:
  """
  Operator to clone a git repository at a given path, usually wherever this operator is called from
  returns True if the clone was successful
  """
  command = ["git", "clone"]
  if branch:
    command.append("-b")
    command.append(branch)
  if depth:
    command.append("--depth")
    command.append(str(depth))
  for k,v in kwargs.items():
    command.append(f"--{k}")
    if v != None:
      command.append(v)
  command.append(repository)
  command.append(path)
  out = Popen(command, stdout=sys.stdout, stderr=sys.stderr).wait()
  return out == 0



"""
{% CallOut variant="warning" label="This is experimental and not ready for production use, can be removed without notice" /%}
"""

# bringing back the dead from the ashes
# this is a file that contains some pre built templates for the users to make their own training workflows

import json
from typing import Dict, Any, Union
from abc import ABCMeta, abstractmethod

from nbox.utils import logger
from nbox.projects import Project
from nbox.sublime.proto.lmao_v2_pb2 import Run as _Runpb


class ModelTrainer(metaclass = ABCMeta):
  def __init__(self, project_id: str):
    self.pid = project_id
    self.proj = None

  def init(self):
    self.proj = Project(id = self.pid)
    self.relic = self.proj.get_relic()

  # There's a couple of things that user needs to fill up in order for this to work smoothely
  # like how does th user want to train the model, or what happens when a user calls to make
  # a inference and saving and loading from checkpoints to quickly get to a certain artifacts

  @abstractmethod
  def train_model(self, *args, **kwargs):
    """
    Trains the model
    """

  @abstractmethod
  def infer(self, *args, **kwargs) -> Dict[str, Any]:
    """
    Runs the inference for any inputs that the user passes
    """

  @abstractmethod
  def save_checkpoint(self, *args, **kwargs) -> str:
    """
    Returns the path to the checkpoint
    """

  @abstractmethod
  def load_checkpoint(self, folder: str) -> None:
    """
    Loads the checkpoint and assigns the relevant attributes like model
    """

  # The next set of functions are provided by this framework to coordinate with NBX to get things
  # done correctly like actual saving of files on the Relics or deploying and running as workflow
  # or logging of metrics

  def log(self, log: Dict[str, Union[str, int, float]], step: int, **kwargs):
    """Logs the metrics to NBX"""
    if hasattr(self, "exp_tracker"):
      self.exp_tracker.log(log, step=step, **kwargs)
    elif hasattr(self, "live_tracker"):
      self.live_tracker.log(log)
    else:
      raise ValueError("No tracker found, initialize using train() or serve()")

  def checkpoint(self, *args, **kwargs):
    save_path = self.save_checkpoint(*args, **kwargs)
    if not type(save_path) == str:
      raise TypeError("Checkpoint path must be a string, save_checkpoint() returned a non string value")
    if not save_path:
      raise ValueError("Checkpoint path cannot be empty, save_checkpoint() returned an empty string")
    logger.info(f"Checkpoint created at {save_path}")
    self.exp_tracker.save_file(save_path)
    self.run_pb = self._update_run_config(
      latest_checkpoint = save_path
    )

  def get_checkpoint(self, experiment_id: str, name: str = ""):
    # if name is not provided we will get whatever is the latest checkpoint
    lmao_stub = self.proj.get_lmao_stub()
    run = lmao_stub.get_run_details(_Runpb(
      workspace_id = self.proj.workspace_id,
      project_id = self.pid,
      experiment_id = experiment_id,
    ))
    if not name:
      name = json.loads(run.config).get("latest_checkpoint", None)
      if name == None:
        raise ValueError("No checkpoint found")
    rp = f'{experiment_id}/{name}'
    logger.info(f"Loading checkpoint:\n  from: {rp}\n  to: {name}")
    self.relic.get_from(name, rp)
    self.load_checkpoint(name)

  def deploy(self, job_id: str):
    pass

  # all the methods below are helpers for the above methods

  def train(self, *args, **kwargs):
    """Run the training part of the pipeline"""
    if self.proj is None:
      self.init()
    self.exp_tracker = self.proj.get_exp_tracker()
    self.run_pb = self.exp_tracker.run
    self.train_model(*args, **kwargs)

  def serve(self, experiment_id: str, checkpoint: str = ""):
    """Run the inference part of the pipeline"""
    if self.proj is None:
      self.init()
    self.live_tracker = self.proj.get_live_tracker()
    self.serving_pb = self.live_tracker.serving
    self.get_checkpoint(experiment_id, checkpoint)

  def _update_run_config(self, **kwargs) -> _Runpb:
    """Updates the run config with the key and value"""
    run = self.run_pb
    cfg = json.loads(run.config)
    for k, v in kwargs.items():
      cfg[k] = v
    run.config = json.dumps(cfg)
    run.update_keys.append("config")
    ack = self.exp_tracker.lmao.update_run_status(run)
    if not ack.success:
      raise Exception(f"Failed to update run config: {ack.message}")
    run.update_keys.pop()
    return run
